import torch
import torch.nn as nn
from .common import *

def sla(x, step):
    diff = x % step
    x += (diff > 0) * (step - diff)
    return torch.arange(x).reshape((-1, step))[:, 0]

# turns a list of layers into a residual block
class ResidualSequential(nn.Sequential):
    def __init__(self, *args):
        super(ResidualSequential, self).__init__(*args)

    def forward(self, x):
        out = super(ResidualSequential, self).forward(x)
        x_ = None
        if out.size(2) != x.size(2) or out.size(3) != x.size(3):
            # x_ = x[:, :, 0 : : 2, 0 : : 2]
            # does the same thing as above but visualization tool HiddenLayer does not support slicing
            x_ = x[:, :, sla(x.shape[2], 2), :][:, :, :, sla(x.shape[3], 2)]
        else:
            x_ = x
        return out + x_

# building block: pre-activation and post-activation
def get_block(in_channels, out_channels, act_fun = 'LeakyReLU', pad = 'reflection', encode = True):
    layers = []

    # pre-activation
    # if encode:
    #     layers = [
    #         bn(in_channels),
    #         act(act_fun),
    #         conv(in_f = in_channels, out_f = out_channels, kernel_size = 3, stride = 2 if encode else 1, bias = True, pad = pad)
    #     ]
    # layers += [
    #     bn(out_channels),
    #     act(act_fun),
    #     conv(in_f = out_channels, out_f = out_channels, kernel_size = 3, stride = 1, bias = True, pad = pad)
    # ]

    # post-activation
    if encode:
        layers = [
            conv(in_f = in_channels, out_f = out_channels, kernel_size = 3, stride = 2, bias = True, pad = pad),
            bn(out_channels),
            act(act_fun)
        ]
    layers += [
        conv(in_f = out_channels, out_f = out_channels, kernel_size = 3, stride = 1, bias = True, pad = pad),
        bn(out_channels),
        act(act_fun)
    ]
    return layers

# Residual U-net
class NewUNet(nn.Module):
    def __init__(self, num_input_channels = 3, num_output_channels = 3,
                 num_channels_down = [16, 32, 64, 128, 128],
                 num_channels_up = [16, 32, 64, 128, 128],
                 num_channels_skip = [4, 4, 4, 4, 4],
                 filter_size_down = 3, filter_size_up = 3, filter_size_skip = 1,
                 upsample_mode = 'bilinear', downsample_mode = 'max',
                 pad = 'relection', act_fun = 'LeakyReLU',
                 need_sigmoid = True, need_bias = True,
                 need1x1_up = True, need_residual = True, need_dropout = True):
        
        super(NewUNet, self).__init__()
        self.dropout_prob = 0.1

        # process arguments
        assert len(num_channels_down) == len(num_channels_up) == len(num_channels_skip)
        n_scales = len(num_channels_down)
        input_depth = num_input_channels

        if not (isinstance(filter_size_down, list) or isinstance(filter_size_down, tuple)):
            filter_size_down = [filter_size_down] * n_scales
        if not (isinstance(filter_size_up, list) or isinstance(filter_size_up, tuple)):
            filter_size_up = [filter_size_up] * n_scales
        if not (isinstance(filter_size_skip, list) or isinstance(filter_size_skip, tuple)):
            filter_size_skip = [filter_size_skip] * n_scales
        if not (isinstance(upsample_mode, list) or isinstance(upsample_mode, tuple)):
            upsample_mode = [upsample_mode] * n_scales
        if not (isinstance(downsample_mode, list) or isinstance(downsample_mode, tuple)):
            downsample_mode = [downsample_mode] * n_scales

        # initialize empty model
        self.model = nn.Sequential()
        model_tmp = self.model
        if need_residual:
            s = ResidualSequential
        else:
            s = nn.Sequential

        for i in range(n_scales):

            # skip
            # conv -> bn -> act
            # channels: input_depth -> num_channels_skip[i]
            # img size: S -> S
            skip = nn.Sequential()
            if num_channels_skip[i] != 0:
                skip.add(conv(input_depth, num_channels_skip[i], filter_size_skip[i], bias = need_bias, pad = pad))
                skip.add(bn(num_channels_skip[i]))
                skip.add(act(act_fun))
                if need_dropout:
                    skip.add(nn.Dropout(p = self.dropout_prob))

            # deeper
            # conv -> bn -> act -> conv -> bn -> act -> downsample
            # channels: input_depth -> num_channels_up[i + 1] if i < last_depth else num_channels_down[i]
            # img size: S -> S/2 -> S
            deeper = nn.Sequential()
            if input_depth != num_channels_down[i]:
                deeper.add(nn.Sequential(*get_block(in_channels = input_depth, out_channels = num_channels_down[i], act_fun = act_fun, pad = pad)))
            else:
                deeper.add(s(*get_block(in_channels = input_depth, out_channels = num_channels_down[i], act_fun = act_fun, pad = pad)))
            if need_dropout:
                deeper.add(nn.Dropout(p = self.dropout_prob))

            # deeper_main
            # deeper building blocks
            # channels: num_channels_down[i] -> k
            # img size: S -> S/2
            deeper_main = nn.Sequential()
            if i == n_scales - 1:
                k = num_channels_down[i]
            else:
                deeper.add(deeper_main)
                k = num_channels_up[i + 1]

            deeper.add(nn.Upsample(scale_factor = 2, mode = upsample_mode[i]))

            # concat
            # concat -> bn -> conv -> bn -> act
            # channels: num_channels_skip[i] + k -> num_channels_up[i]
            if num_channels_skip[i] != 0:
                model_tmp.add(Concat(1, skip, deeper))
            else:
                model_tmp.add(deeper)
                # model_tmp.add(bn(num_channels_skip[i] + k))
            model_tmp.add(s(*get_block(in_channels = num_channels_skip[i] + k, out_channels = num_channels_up[i], act_fun = act_fun, pad = pad, encode = False)))
            if need_dropout:
                model_tmp.add(nn.Dropout(p = self.dropout_prob))

            # 1x1 convolution block
            if need1x1_up:
                model_tmp.add(conv(num_channels_up[i], num_channels_up[i], 1, bias = need_bias, pad = pad))
                model_tmp.add(bn(num_channels_up[i]))
                model_tmp.add(act(act_fun))
                if need_dropout:
                    model_tmp.add(nn.Dropout(p = self.dropout_prob))

            input_depth = num_channels_down[i]
            model_tmp = deeper_main

        # output layer
        self.model.add(conv(num_channels_up[0], num_output_channels, 1, bias = need_bias, pad = pad))
        if need_sigmoid:
            self.model.add(nn.Sigmoid())

    def forward(self, input):
        return self.model(input)

    def eval(self):
        self.model.eval()
