from __future__ import print_function
import matplotlib.pyplot as plt
import numpy as np
import os
from skimage.metrics import peak_signal_noise_ratio as compare_psnr
import time
import torch
import torch.optim
import torchsummary
import hiddenlayer

from utils.denoising_utils import *
from models import *

output_folder = 'results/'
torch.backends.cudnn.enabled = True
torch.backends.cudnn.benchmark =True

# This model in FloatTensor converges on k80 but not on V100
# please use DoubleTensor if not converge
# dtype = torch.cuda.FloatTensor
dtype = torch.cuda.DoubleTensor
num_iter = 5000
LR = 0.01

imsize =-1
PLOT = True
sigma = 25
sigma_ = sigma/255.

# load image and add noise
fname = 'data/denoising/F16_GT.png'
img_pil = crop_image(get_image(fname, imsize)[0], d=32)
img_np = pil_to_np(img_pil)
img_noisy_pil, img_noisy_np = get_noisy_image(img_np, sigma_)
img_noisy_torch = np_to_torch(img_noisy_np).type(dtype)
if PLOT:
    plot_image_grid([img_np, img_noisy_np], 4, 6);
    plt.savefig(output_folder + 'truth-noisy.jpg')

# setup parameters
INPUT = 'noise' # 'meshgrid'
pad = 'reflection'
OPT_OVER = 'net' # 'net,input'
reg_noise_std = 1./30. # set to 1./20. for sigma=50
OPTIMIZER='adam' # 'LBFGS'
show_every = 100
exp_weight=0.99
input_depth = 32 
figsize = 4

# create net
# ResNet
# net = get_net(input_depth, 'ResNet', pad = pad, upsample_mode = 'bilinear').type(dtype)
# U-net
# net = get_net(input_depth, 'skip', pad,
#               skip_n33d=128, 
#               skip_n33u=128, 
#               skip_n11=4, 
#               num_scales=5,
#               upsample_mode='bilinear').type(dtype)
# Residual U-Net
net = get_net(input_depth, 'NewUNet', pad,
              skip_n33d = 108,
              skip_n33u = [128, 124, 120, 116, 112],
              skip_n11 = 4,
              num_scales = 5,
              upsample_mode = 'bilinear').type(dtype)
# number of parameters
s  = sum([np.prod(list(p.size())) for p in net.parameters()]); 
print ('Number of params: %d' % s)

net.eval()

# visualize model
torchsummary.summary(net.type(torch.cuda.FloatTensor), input_size = (input_depth, 512, 512))
transforms = [hiddenlayer.transforms.Prune('Constant')]
graph = hiddenlayer.build_graph(net.type(torch.FloatTensor), torch.zeros(1, input_depth, 512, 512), transforms = transforms)
graph.theme = hiddenlayer.graph.THEMES['blue'].copy()
graph.save(output_folder + 'model', format = 'png')


# Loss
mse = torch.nn.MSELoss().type(dtype)
net.train()
net = net.type(dtype)

# setup booking variables
net_input = get_noise(input_depth, INPUT, (img_pil.size[1], img_pil.size[0])).type(dtype).detach()

net_input_saved = net_input.detach().clone()
noise = net_input.detach().clone()
out_avg = None
last_net = None
max_out_avg_psnr = 0
max_out_avg = None
psrn_noisy_last = 0

psnrs = [[], [], []]

i = 0
def closure():
    
    global i, out_avg, psrn_noisy_last, last_net, net_input, max_out_avg_psnr, max_out_avg
    
    if reg_noise_std > 0:
        net_input = net_input_saved + (noise.normal_() * reg_noise_std)
    
    out = net(net_input)
    
    # Smoothing
    if out_avg is None:
        out_avg = out.detach()
    else:
        out_avg = out_avg * exp_weight + out.detach() * (1 - exp_weight)
            
    total_loss = mse(out, img_noisy_torch)
    total_loss.backward()
        
    psrn_noisy = compare_psnr(img_noisy_np, out.detach().cpu().numpy()[0]) 
    psrn_gt    = compare_psnr(img_np, out.detach().cpu().numpy()[0]) 
    psrn_gt_sm = compare_psnr(img_np, out_avg.detach().cpu().numpy()[0])
    if psrn_gt_sm > max_out_avg_psnr:
        max_out_avg_psnr = psrn_gt_sm
        max_out_avg = out_avg

    psnrs[0].append(psrn_noisy)
    psnrs[1].append(psrn_gt)
    psnrs[2].append(psrn_gt_sm)
    
    # Note that we do not have GT for the "snail" example
    # So 'PSRN_gt', 'PSNR_gt_sm' make no sense
    print('Iteration %05d\tLoss %f\tPSNR_noisy %f\tPSNR_gt %f\tPSNR_gt_sm %f' % (i, total_loss.item(), psrn_noisy, psrn_gt, psrn_gt_sm))
    if  PLOT and i % show_every == 0:
        out_np = torch_to_np(out.type(torch.cuda.FloatTensor))
        out_avg_np = torch_to_np(out_avg.type(torch.cuda.FloatTensor))
        plot_image_grid([img_noisy_np,
                         np.clip(out_np, 0, 1), 
                         np.clip(out_avg_np, 0, 1)], factor=figsize, nrow=1)
        plt.savefig(output_folder + 'iteration' + str(i) + '.jpg')
    
    # Backtracking
    if i % show_every:
        if psrn_noisy - psrn_noisy_last < -5: 
            print('Falling back to previous checkpoint.')

            for new_param, net_param in zip(last_net, net.parameters()):
                net_param.data.copy_(new_param.cuda())

            return total_loss*0
        else:
            last_net = [x.detach().cpu() for x in net.parameters()]
            psrn_noisy_last = psrn_noisy
    
    i += 1

    return total_loss

# train
p = get_params(OPT_OVER, net, net_input)
start = time.time()
optimize(OPTIMIZER, p, closure, LR, num_iter)
print("Denoising takes %f seconds" % (time.time() - start))

# plot PSNR curve
plt.figure()
plt.plot(psnrs[0])
plt.plot(psnrs[1])
plt.plot(psnrs[2])
plt.legend(['noisy', 'truth', 'avg-truth'])
plt.xlabel('Iterations')
plt.ylabel('PSNR')
plt.savefig(output_folder + 'training_curve.jpg')

# plot final results
net.eval()
out_np = torch_to_np(net(net_input).type(torch.cuda.FloatTensor))
out_avg_np = torch_to_np(max_out_avg.type(torch.cuda.FloatTensor))
q = plot_image_grid([img_np, np.clip(out_np, 0, 1), np.clip(out_avg_np, 0, 1)], factor=13);
plt.savefig(output_folder + '/final results.jpg')
print("Final one pass PSNR =", compare_psnr(img_np, out_np))
print("Max averaged image PSNR =", max_out_avg_psnr)
